@@ -28,7 +28,7 @@ local config_schema = {
         apisix = {
             properties = {
                 config_center = {
-                    enum = {"etcd", "yaml"},
+                    enum = {"etcd", "yaml", "xds"},
                 },
                 lua_module_hook = {
                     pattern = "^[a-zA-Z._-]+$",
@@ -212,19 +212,8 @@ local config_schema = {
                             type = "string",
                         },
                     }
-                },
-                prefix = {
-                    type = "string",
-                },
-                host = {
-                    type = "array",
-                    items = {
-                        type = "string",
-                        pattern = [[^https?://]]
-                    }
                 }
-            },
-            required = {"prefix", "host"}
+            }
         },
         wasm = {
             type = "object",
