@@ -38,19 +38,20 @@ local apisix_upstream = require("apisix.upstream")
 local set_upstream    = apisix_upstream.set_by_route
 local apisix_ssl      = require("apisix.ssl")
 local upstream_util   = require("apisix.utils.upstream")
+local xrpc            = require("apisix.stream.xrpc")
 local ctxdump         = require("resty.ctxdump")
-local ipmatcher       = require("resty.ipmatcher")
 local ngx_balancer    = require("ngx.balancer")
 local debug           = require("apisix.debug")
+local pubsub_kafka    = require("apisix.pubsub.kafka")
 local ngx             = ngx
 local get_method      = ngx.req.get_method
 local ngx_exit        = ngx.exit
 local math            = math
 local error           = error
 local ipairs          = ipairs
-local tostring        = tostring
 local ngx_now         = ngx.now
 local ngx_var         = ngx.var
+local re_split        = require("ngx.re").split
 local str_byte        = string.byte
 local str_sub         = string.sub
 local tonumber        = tonumber
@@ -92,6 +93,8 @@ function _M.http_init(args)
             core.log.error("failed to load the configuration: ", err)
         end
     end
+
+    xrpc.init()
 end
 
 
@@ -122,16 +125,20 @@ function _M.http_init_worker()
 
     require("apisix.debug").init_worker()
 
+    if core.config.init_worker then
+        local ok, err = core.config.init_worker()
+        if not ok then
+            core.log.error("failed to init worker process of ", core.config.type,
+                           " config center, err: ", err)
+        end
+    end
+
     plugin.init_worker()
     router.http_init_worker()
     require("apisix.http.service").init_worker()
     plugin_config.init_worker()
     require("apisix.consumer").init_worker()
 
-    if core.config == require("apisix.core.config_yaml") then
-        core.config.init_worker()
-    end
-
     apisix_upstream.init_worker()
     require("apisix.plugins.ext-plugin.init").init_worker()
 
@@ -167,61 +174,9 @@ function _M.http_ssl_phase()
 end
 
 
-
-
-local function parse_domain_for_nodes(nodes)
-    local new_nodes = core.table.new(#nodes, 0)
-    for _, node in ipairs(nodes) do
-        local host = node.host
-        if not ipmatcher.parse_ipv4(host) and
-                not ipmatcher.parse_ipv6(host) then
-            local ip, err = core.resolver.parse_domain(host)
-            if ip then
-                local new_node = core.table.clone(node)
-                new_node.host = ip
-                new_node.domain = host
-                core.table.insert(new_nodes, new_node)
-            end
-
-            if err then
-                core.log.error("dns resolver domain: ", host, " error: ", err)
-            end
-        else
-            core.table.insert(new_nodes, node)
-        end
-    end
-    return new_nodes
-end
-
-
-local function parse_domain_in_up(up)
-    local nodes = up.value.nodes
-    local new_nodes, err = parse_domain_for_nodes(nodes)
-    if not new_nodes then
-        return nil, err
-    end
-
-    local ok = upstream_util.compare_upstream_node(up.dns_value, new_nodes)
-    if ok then
-        return up
-    end
-
-    if not up.orig_modifiedIndex then
-        up.orig_modifiedIndex = up.modifiedIndex
-    end
-    up.modifiedIndex = up.orig_modifiedIndex .. "#" .. ngx_now()
-
-    up.dns_value = core.table.clone(up.value)
-    up.dns_value.nodes = new_nodes
-    core.log.info("resolve upstream which contain domain: ",
-                  core.json.delay_encode(up, true))
-    return up
-end
-
-
 local function parse_domain_in_route(route)
     local nodes = route.value.upstream.nodes
-    local new_nodes, err = parse_domain_for_nodes(nodes)
+    local new_nodes, err = upstream_util.parse_domain_for_nodes(nodes)
     if not new_nodes then
         return nil, err
     end
@@ -263,11 +218,8 @@ local function set_upstream_host(api_ctx, picked_server)
     end
 
     local nodes_count = up_conf.nodes and #up_conf.nodes or 0
-    if nodes_count == 1 then
-        local node = up_conf.nodes[1]
-        api_ctx.var.upstream_host = node.domain or node.host
-    elseif picked_server.domain and ngx_balancer.recreate_request then
-        api_ctx.var.upstream_host = picked_server.domain
+    if nodes_count == 1 or ngx_balancer.recreate_request then
+        api_ctx.var.upstream_host = picked_server.upstream_host
     end
 end
 
@@ -275,41 +227,9 @@ end
 local function set_upstream_headers(api_ctx, picked_server)
     set_upstream_host(api_ctx, picked_server)
 
-    local proto = api_ctx.var.http_x_forwarded_proto
-    if proto then
-        api_ctx.var.var_x_forwarded_proto = proto
-    end
-end
-
-
-local function get_upstream_by_id(up_id)
-    local upstreams = core.config.fetch_created_obj("/upstreams")
-    if upstreams then
-        local upstream = upstreams:get(tostring(up_id))
-        if not upstream then
-            core.log.error("failed to find upstream by id: " .. up_id)
-            if is_http then
-                return core.response.exit(502)
-            end
-
-            return ngx_exit(1)
-        end
-
-        if upstream.has_domain then
-            local err
-            upstream, err = parse_domain_in_up(upstream)
-            if err then
-                core.log.error("failed to get resolved upstream: ", err)
-                if is_http then
-                    return core.response.exit(500)
-                end
-
-                return ngx_exit(1)
-            end
-        end
-
-        core.log.info("parsed upstream: ", core.json.delay_encode(upstream, true))
-        return upstream.dns_value or upstream.value
+    local hdr = core.request.header(api_ctx, "X-Forwarded-Proto")
+    if hdr then
+        api_ctx.var.var_x_forwarded_proto = hdr
     end
 end
 
@@ -344,6 +264,44 @@ local function verify_tls_client(ctx)
 end
 
 
+local function normalize_uri_like_servlet(uri)
+    local found = core.string.find(uri, ';')
+    if not found then
+        return uri
+    end
+
+    local segs, err = re_split(uri, "/", "jo")
+    if not segs then
+        return nil, err
+    end
+
+    local len = #segs
+    for i = 1, len do
+        local seg = segs[i]
+        local pos = core.string.find(seg, ';')
+        if pos then
+            seg = seg:sub(1, pos - 1)
+            -- reject bad uri which bypasses with ';'
+            if seg == "." or seg == ".." then
+                return nil, "dot segment with parameter"
+            end
+            if seg == "" and i < len then
+                return nil, "empty segment with parameters"
+            end
+
+            segs[i] = seg
+
+            seg = seg:lower()
+            if seg == "%2e" or seg == "%2e%2e" then
+                return nil, "encoded dot segment"
+            end
+        end
+    end
+
+    return core.table.concat(segs, '/')
+end
+
+
 local function common_phase(phase_name)
     local api_ctx = ngx.ctx.api_ctx
     if not api_ctx then
@@ -377,11 +335,25 @@ function _M.http_access_phase()
     debug.dynamic_debug(api_ctx)
 
     local uri = api_ctx.var.uri
-    if local_conf.apisix and local_conf.apisix.delete_uri_tail_slash then
-        if str_byte(uri, #uri) == str_byte("/") then
-            api_ctx.var.uri = str_sub(api_ctx.var.uri, 1, #uri - 1)
-            core.log.info("remove the end of uri '/', current uri: ",
-                          api_ctx.var.uri)
+    if local_conf.apisix then
+        if local_conf.apisix.delete_uri_tail_slash then
+            if str_byte(uri, #uri) == str_byte("/") then
+                api_ctx.var.uri = str_sub(api_ctx.var.uri, 1, #uri - 1)
+                core.log.info("remove the end of uri '/', current uri: ", api_ctx.var.uri)
+            end
+        end
+
+        if local_conf.apisix.normalize_uri_like_servlet then
+            local new_uri, err = normalize_uri_like_servlet(uri)
+            if not new_uri then
+                core.log.error("failed to normalize: ", err)
+                return core.response.exit(400)
+            end
+
+            api_ctx.var.uri = new_uri
+            -- forward the original uri so the servlet upstream
+            -- can consume the param after ';'
+            api_ctx.var.upstream_uri = uri
         end
     end
 
@@ -447,10 +419,6 @@ function _M.http_access_phase()
     api_ctx.route_id = route.value.id
     api_ctx.route_name = route.value.name
 
-    local ref = ctxdump.stash_ngx_ctx()
-    core.log.info("stash ngx ctx: ", ref)
-    ngx_var.ctx_ref = ref
-
     -- run global rule
     plugin.run_global_rules(api_ctx, router.global_rules, nil)
 
@@ -493,7 +461,15 @@ function _M.http_access_phase()
     end
 
     if up_id then
-        local upstream = get_upstream_by_id(up_id)
+        local upstream = apisix_upstream.get_by_id(up_id)
+        if not upstream then
+            if is_http then
+                return core.response.exit(502)
+            end
+
+            return ngx_exit(1)
+        end
+
         api_ctx.matched_upstream = upstream
 
     else
@@ -529,6 +505,13 @@ function _M.http_access_phase()
         api_ctx.upstream_scheme = "grpc"
     end
 
+    -- load balancer is not required by kafka upstream, so the upstream
+    -- node selection process is intercepted and left to kafka to
+    -- handle on its own
+    if api_ctx.matched_upstream and api_ctx.matched_upstream.scheme == "kafka" then
+        return pubsub_kafka.access(api_ctx)
+    end
+
     local code, err = set_upstream(route, api_ctx)
     if code then
         core.log.error("failed to set upstream: ", err)
@@ -548,6 +531,10 @@ function _M.http_access_phase()
     -- run the before_proxy method in access phase first to avoid always reinit request
     common_phase("before_proxy")
 
+    local ref = ctxdump.stash_ngx_ctx()
+    core.log.info("stash ngx ctx: ", ref)
+    ngx_var.ctx_ref = ref
+
     local up_scheme = api_ctx.upstream_scheme
     if up_scheme == "grpcs" or up_scheme == "grpc" then
         return ngx.exec("@grpc_pass")
@@ -847,6 +834,8 @@ function _M.stream_init(args)
             core.log.error("failed to load the configuration: ", err)
         end
     end
+
+    xrpc.init()
 end
 
 
@@ -862,6 +851,7 @@ function _M.stream_init_worker()
     core.log.info("random stream test in [1, 10000]: ", math.random(1, 10000))
 
     plugin.init_worker()
+    xrpc.init_worker()
     router.stream_init_worker()
     apisix_upstream.init_worker()
 
@@ -876,15 +866,9 @@ end
 
 
 function _M.stream_preread_phase()
-    core.log.info("enter stream_preread_phase")
-
     local ngx_ctx = ngx.ctx
-    local api_ctx = ngx_ctx.api_ctx
-
-    if not api_ctx then
-        api_ctx = core.tablepool.fetch("api_ctx", 0, 32)
-        ngx_ctx.api_ctx = api_ctx
-    end
+    local api_ctx = core.tablepool.fetch("api_ctx", 0, 32)
+    ngx_ctx.api_ctx = api_ctx
 
     if not verify_tls_client(api_ctx) then
         return ngx_exit(1)
@@ -909,7 +893,17 @@ function _M.stream_preread_phase()
 
     local up_id = matched_route.value.upstream_id
     if up_id then
-        api_ctx.matched_upstream = get_upstream_by_id(up_id)
+        local upstream = apisix_upstream.get_by_id(up_id)
+        if not upstream then
+            if is_http then
+                return core.response.exit(502)
+            end
+
+            return ngx_exit(1)
+        end
+
+        api_ctx.matched_upstream = upstream
+
     else
         if matched_route.has_domain then
             local err
@@ -938,6 +932,11 @@ function _M.stream_preread_phase()
 
     plugin.run_plugin("preread", plugins, api_ctx)
 
+    if matched_route.value.protocol then
+        xrpc.run_protocol(matched_route.value.protocol, api_ctx)
+        return
+    end
+
     local code, err = set_upstream(matched_route, api_ctx)
     if code then
         core.log.error("failed to set upstream: ", err)
