@@ -20,6 +20,8 @@ local plugin_name = "api-breaker"
 local ngx = ngx
 local math = math
 local error = error
+local ipairs = ipairs
+
 
 local shared_buffer = ngx.shared["plugin-".. plugin_name]
 if not shared_buffer then
@@ -35,6 +37,25 @@ local schema = {
             minimum = 200,
             maximum = 599,
         },
+        break_response_body = {
+            type = "string"
+        },
+        break_response_headers = {
+            type = "array",
+            items = {
+                type = "object",
+                properties = {
+                    key = {
+                        type = "string",
+                        minLength = 1
+                    },
+                    value = {
+                        type = "string",
+                        minLength = 1
+                    }
+                }
+            }
+        },
         max_breaker_sec = {
             type = "integer",
             minimum = 3,
@@ -158,6 +179,15 @@ function _M.access(conf, ctx)
 
     -- breaker
     if lasttime + breaker_time >= ngx.time() then
+        if conf.break_response_body then
+            if conf.break_response_headers then
+                for _, value in ipairs(conf.break_response_headers) do
+                    local val = core.utils.resolve_var(value.value, ctx.var)
+                    core.response.add_header(value.key, val)
+                end
+            end
+            return conf.break_response_code, conf.break_response_body
+        end
         return conf.break_response_code
     end
 
