@@ -45,7 +45,8 @@ local schema = {
             }
         },
         concat_method = {type = "string", default = "json",
-                         enum = {"json", "new_line"}}
+                         enum = {"json", "new_line"}},
+        ssl_verify = {type = "boolean", default = false},
     },
     required = {"uri"}
 }
@@ -106,7 +107,7 @@ local function send_http_data(conf, log_message)
     end
 
     if url_decoded.scheme == "https" then
-        ok, err = httpc:ssl_handshake(true, host, false)
+        ok, err = httpc:ssl_handshake(true, host, conf.ssl_verify)
         if not ok then
             return false, "failed to perform SSL with host[" .. host .. "] "
                 .. "port[" .. tostring(port) .. "] " .. err
